/*****************************************************************************

LOADER_FS_GF rev 0.03

FS = File System 2
GF = gForth


IMPORTANT: be sure "Include debug code RST/28 instructions" is NOT checked

then <Compile> "Compile To Target" "Compile to Flash"

  Run with a serial utility such as Kermit.
  Configure the serial utility for port connected to RS232, 19200 8N1.



 10/20/01	Initial coding.
 10/24/01	Changed assy parameter for baud & result to HL
 10/25/01	Hard coded flash buffer for extension code
 01/26/02	Test MMU remap after xmem2xmem
 01/29/02	forth now runs with fully mapped ram at $0000
 				No BIOS customization needed.
 04/15/03	Serial port selection via define
 				Better dialog on flash exit
 06/03/03	Changed to FS2 in program Flash for image (was WriteFlash)
 07/02/04   Fix assy use of vars must be static (DC8.x changed default)
 				Started versioning file names.
 07/08/04	Rev 0.02: Changed RAM allocation to 56K.
 07/09/04	Rev 0.03: fixed static addr; fixed multi-line macros
 				added ErrorHandler;
 				increased FS size - edit FS_DEV.LIB (ver 7.21 and earlier)
 										- edit RABBITBIOS.C XMEM_RESERVE

*****************************************************************************/

#define TRUE 1
#define FALSE 0

#define DEBUG_PRINT FALSE		// ENABLE stdio print?

enum errnum {
	err_fwrite,
   err_LX_dev,
   err_fsinit,
   err_format,
   err_fcreate,
   err_formatwrite,
   err_alloc,
   err_imagelength,
   err_fcreatec,
   err_xmem2root,
   err_fopenc,
   err_reload,
   err_fcreatex,
   err_fopenX,
   err_xmem2rootB
   };

#if DEBUG_PRINT == TRUE
#define PR_DEBUG(string)  printf( string ) ;
#else
#define PR_DEBUG(string)
#endif

#if DEBUG_PRINT == TRUE
#define PR_DEBUG1(string,arg)  printf( string , arg ) ;
#else
#define PR_DEBUG1(string,arg)
#endif


#define PR_SERIAL(string) {\
      SerialPuts( string );\
      while (SerialwrFree() != OUTBUFSIZE) ; }

#define PR_SERIAL1(string,arg) {\
		sprintf(G_buf, string, arg);\
      SerialPuts( G_buf );\
      while (SerialwrFree() != OUTBUFSIZE) ; }



// FS2 seems to need up to 2x the space actually used...

// Edit BIOS\RABBITBIOS.C : XMEM_RESERVE_SIZE = 0x10000L	// 64KB


#define FS2_USE_PROGRAM_FLASH	64	// KB

#define LX_2_USE			fs_get_flash_lx()

#define MY_LS_SHIFT		8	// Log(base 2) of the desired LS size.  Run the
									// FS2INFO sample program to check out
									// suitable values.  The LS (logical sector)
									// size is the internal buffer unit used
									// by the filesystem.

#define FS_MAX_FILES		4		// Maximum files in filesystem


#if DEBUG_PRINT == TRUE
#define FS_API_TRACE				// Open/close/create/delete tracing to stdout.
#else
#undef FS_API_TRACE
#endif



#use "fs2.lib"





// use xmem for main()!!
//
// it is critical that the in-line
// assembly that reconfigures the MMU and calls FORTH
// be located in xmem (0xE000-0xFFFF) so it is still mapped
// during the flash to ram and back change
//

#memmap xmem


/*****************************************************************************

 DEFINES to make it easier to change the serial port

 Change the definition of PORT in this file
 _AND_
 change the definition of PORT in the assembly source file

*****************************************************************************/

#define spA 1
#define spB 2
#define spC 3

#define PORT spA


// go ahead and define buffer sizes for all possible serial ports

#define AINBUFSIZE 15
#define BINBUFSIZE 15
#define CINBUFSIZE 15

#define OUTBUFSIZE 63

#define AOUTBUFSIZE 63
#define BOUTBUFSIZE 63
#define COUTBUFSIZE 63



#define BAUDRATE 19200


/*****************************************************************************/


#ximport "gfr2k.bin" forthbin


// places the length of <filename> ( stored as a long)
// and its binary contents at the next available place in xmem flash.
// forthbin is a 'constant MACRO' that resolves to a long (address).

// gforth is about 18k bytes.


/*****************************************************************************/




#define FORTHORG 0x0000

// forth is in assy. org to start at 0x0000 (in RAM).



// GLOBAL BUFFER for serial output

char G_buf[80];



/*****************************************************************************

SERIAL wrapper routines

*****************************************************************************/

int SerialOpen(long br)
{
#if PORT==spA
  return(serAopen(br));
#elif PORT==spB
  return(serBopen(br));
#elif PORT==spC
  return(serCopen(br));
#endif

}

int SerialPutc(char ch)
{
#if PORT==spA
  return(serAputc(ch));
#elif PORT==spB
  return(serBputc(ch));
#elif PORT==spC
  return(serCputc(ch));
#endif

}

int SerialGetc()
{
#if PORT==spA
  return(serAgetc());
#elif PORT==spB
  return(serBgetc());
#elif PORT==spC
  return(serCgetc());
#endif

}

int SerialPuts(char *str)
{
#if PORT==spA
  return(serAputs(str));
#elif PORT==spB
  return(serBputs(str));
#elif PORT==spC
  return(serCputs(str));
#endif

}

int SerialwrFree()
{
#if PORT==spA
  return(serAwrFree());
#elif PORT==spB
  return(serBwrFree());
#elif PORT==spC
  return(serCwrFree());
#endif
}

void SerialClose()
{
#if PORT==spA
  serAclose();
#elif PORT==spB
  serBclose();
#elif PORT==spC
  serCclose();
#endif
}


void ErrorHandler(enum errnum err, int errno){

   SerialOpen(BAUDRATE);

   switch (err){
   case err_fwrite:	PR_SERIAL1("\r\nError: fwrite (%d)\r\n",errno); break;
   case err_LX_dev:  PR_SERIAL1("\r\nError: LX device doesn't exist (%d)\r\n",errno); break;
   case err_fsinit:  PR_SERIAL1("\r\nError: fs_init (%d)\r\n",errno); break;
   case err_format:  PR_SERIAL1("\r\nError: format (%d)\r\n",errno); break;
   case err_fcreate:  PR_SERIAL1("\r\nError: fcreate (%d)\r\n",errno); break;
   case err_formatwrite:  PR_SERIAL1("\r\nError: format test failed (%d)\r\n",errno); break;
   case err_alloc:  PR_SERIAL1("\r\nError: xalloc failed (%d)\r\n",errno); break;
   case err_imagelength:  PR_SERIAL1("\r\nError: illegal image size (%d)\r\n",errno); break;
   case err_fcreatec:  PR_SERIAL1("\r\nError: fcreate during init (%d)\r\n",errno); break;
   case err_xmem2root:  PR_SERIAL1("\r\nError: xmem2root during FS (%d)\r\n",errno); break;
   case err_fopenc:  PR_SERIAL1("\r\nError: fopen (%d)\r\n",errno); break;
   case err_reload:  PR_SERIAL1("\r\nError: bad image. reboot to reload. (%d)\r\n",errno); break;
   case err_fcreatex:  PR_SERIAL1("\r\nError: fcreate during flash exit (%d)\r\n",errno); break;
   case err_fopenX:  PR_SERIAL1("\r\nError: fopen during flash exit (%d)\r\n",errno); break;
   case err_xmem2rootB:  PR_SERIAL1("\r\nError: xmem2root during flash exit (%d)\r\n",errno); break;

   default:          PR_SERIAL("\r\nError: unknown error code!\r\n"); break;
   }


   SerialClose();
   exit(-1);
}

/*****************************************************************************

	Routines for segment read

*****************************************************************************/

int get_dataseg( )
{
#asm nodebug
	ioi ld a,(DATASEG)
	ld l,a
	ld h,0
#endasm
}

int get_stackseg( )
{
#asm nodebug
	ioi ld a,(STACKSEG)
	ld l,a
	ld h,0
#endasm
}

int get_segsize( )
{
#asm nodebug
	ioi ld a,(SEGSIZE)
	ld l,a
	ld h,0
#endasm
}

// (root) data builds down from (watch_code_base - 1)

// keep out of:
// extended memory logical address region E000H to FFFFH



#define READ_SIZE 256



/*****************************************************************************

	Routines for FS

*****************************************************************************/


// reads FS file into xmem buffer

void read_file(File * f, unsigned long addr)
{
	char buffer[READ_SIZE];
	int rc;
	unsigned int offset;

	fseek(f, 0, SEEK_SET);	// start at the beginning

	offset=0;
	do {
		rc = fread(f, buffer, READ_SIZE);
		if (rc)
		  {
			root2xmem((addr+offset), buffer, rc);
			offset=offset+READ_SIZE;
		  }
		} while (rc > 0);
}


// read long data from FS file

long read_long(File * f, int offset)
{
	int rc;
	char buffer[8];
	long result;

	fseek(f, offset ,SEEK_SET);	// start at the beginning

   rc = fread(f, buffer, 4);
   if (rc == 0)
     {
		PR_DEBUG1("read errno = %d\n",errno)
		result = -1;
	  }
	else
	  {
	  result = buffer[0] | ((long)buffer[1]) << 8 | \
	    ((long)buffer[2]) << 16 | ((long)buffer[3]) << 24;
	  }
	return ( result );
}

// fwrite limit is 32767 max. data must be in root space.
// write by appending smaller blocks.

void write_file(File * f, char buffer[], int numbytes)
{
	int rc;
	int ec;


	fseek(f, 0, SEEK_END);	// append

   rc = fwrite(f, buffer, numbytes);
   if (rc != numbytes)
     {
     PR_DEBUG1("write failed! errno = %d\n",errno)
     ErrorHandler(err_fwrite,errno);
     }

}





/*****************************************************************************


							MAIN


*****************************************************************************/

// the stdio printf output code is executed during the
// 1st run of the program while in program mode
// (attached to PC running Dyn-C).

// The binary image is loaded into xmem at compile time.
// On the first run this image is copied into the file system.
// We decide if it is the first run by looking at the file's length location
// and looking at the base+length location (right after the image) in the file,
// which is set to be the same as the length after the first load.


xmem main()
{
	char buf[32];

	int i;
	int c;
	int err;
	int ec;
	int nblks;

   long baud;
   long divisor;

   unsigned long addr1;
   unsigned long addr2;


   unsigned long length;
   unsigned long f_length;
   unsigned long f_check;

   unsigned int len16;

// these 3 vars are referenced by the assembly code that does
// memory re-mapping. they must be static.

   static int br_div;

   static unsigned int segsize_val;
   static unsigned int dataseg_val;

// and assembly code assigns return value...
	static int addr;	// return value from FORTH

   unsigned char nss;
   unsigned char nds;
   unsigned char oss;
   unsigned char ods;

	int ValidReply;


#define BSIZE 0x1000				// 4K buffer

	static unsigned char buffer[BSIZE];		// in root space

// FS2 vars...
	int rc;						// Return code from filesystem functions

	File fh;						// file handle
	FileNumber fhn;			// Corresponding file number


   static unsigned long fpos;


// INITIALIZE some vars...

// INITIALIZE the filesystem...
	if (!LX_2_USE)
	  {
		printf("The specified device (LX# %d) does not exist.  Change LX_2_USE.\n",
			(int)LX_2_USE);
		ErrorHandler(err_LX_dev,((int)LX_2_USE));
	  }
	else
	  {
		printf("Using device LX# %d...\n", (int)LX_2_USE);
	  }

	/* initialize the filesystem. */
	rc = fs_init(0,0);
	if (rc)
	  {
		printf("Could not initialize filesystem, error number %d\n", errno);
		ErrorHandler(err_fsinit,errno);
	  }

   rc = fopen_rd(&fh,2);           // (test file 2?) read handle
   if (rc)
     {
     // this file indicates that the FS has been formated!
     PR_DEBUG("formatting file system!\n")
     rc = lx_format(LX_2_USE, 0);
     if (rc)
       {
       printf("Format failed, error code %d\n", errno);
       ErrorHandler(err_format,errno);
       }

     fs_set_lx(LX_2_USE, LX_2_USE);

     rc = fcreate(&fh,2);      // leaves handle open (!) for writes...
     if (rc)
       {
       PR_DEBUG("failed to create file! errno = \n")
       ErrorHandler(err_fcreate,errno);
       }

     strcpy(buf, "FORMATTED");
     rc = fwrite(&fh, buf, strlen(buf));
     if (rc != strlen(buf))
       {
       printf("formatted indicator file write failed!\n");
       ErrorHandler(err_formatwrite,rc);
       }

     fclose(&fh); // close write handle (from create)
     }
   else
     {
     fclose(&fh); // close test read handle
     }


	fs_set_lx(LX_2_USE, LX_2_USE);

	fhn = 1;		// only one file. always "named" 1.


// allocate 56K + 4K of ram.
// we can't use all of this (because we MUST keep out of
// xmem space 0xE000..0xFFFF). the extra 4K alloc is to make room
// for 4K alignment.

	addr1 = xalloc(0xF000);

	if (addr1 == -1)
	  {
	  PR_DEBUG("failed to allocate RAM!\n")
	  ErrorHandler(err_alloc,-1);
	  }
	else
	  {
     PR_DEBUG1("physical address allocated: %06lx\n",(long)addr1)
	  }


// END OF INITIALIZATION


	xmem2root(&length,forthbin,sizeof(long));		// set length (var in root)
																// for copying image into ram

	if (length > 0xFFFF)
	  {
	  PR_DEBUG("Illegal binary image length!\n")
	  ErrorHandler(err_imagelength,-1);
	  }

	len16 = (int)length;									// not more than 64K by design!
	PR_DEBUG1("length = %04X\n",len16)

	xmem2xmem(addr1,forthbin,(len16+sizeof(long)));			// copy length _and_ image to buffer in ram

	root2xmem((addr1+len16+sizeof(long)), &length, sizeof(long));		// append f_check


// figure out if FS image is valid by looking at the length (first long data)
// and comparing with check (appended to the binary image).
// if FS image is invalid, then reload it from fbi


// does file exist?
	rc = fopen_rd(&fh,fhn);		// (test) read handle

	// if it doesn't then create it! should only run 1st time after format.
	if (rc)
	  {
	  PR_DEBUG("file doesn't exist!\ncreating file\n")

     fs_set_lx(LX_2_USE, LX_2_USE);	  // recommended before create...

	  rc = fcreate(&fh,fhn);	// leaves handle open (!) for writes...
	  if (rc)
	    {
	    PR_DEBUG("failed to create file! errno = \n")
	    ErrorHandler(err_fcreatec,errno);
	    }
	  else	// write to created file...
	    {		//   put initial image in file a buffer full at a time!
 	    nblks = ((len16+(2*sizeof(long)))/BSIZE) + 1;

		 for(i=0;i<nblks;i++)
		   {
         ec = xmem2root(buffer, (addr1 + i*BSIZE), BSIZE);
		   if (ec) ErrorHandler(err_xmem2root,ec);
		   write_file(&fh,buffer,BSIZE);
		   PR_DEBUG1("wrote buffer %d \n", i)
		   }

		 fclose(&fh);	// close write handle from create
	    }
	  }
	else
	  {
	  fclose(&fh);		// close (test) read handle
	  }

   rc = fopen_rd(&fh,fhn);	/* now open it for READ... */
   if (rc)
     {
     PR_DEBUG("problem opening file!\n")
     ErrorHandler(err_fopenc,rc);
     }

	fpos = ftell(&fh);
	PR_DEBUG1("file position should be zero:  %06X\n",fpos)

	// loadFSimage() if length < 0xC000, otherwise reimageFS()!

	f_length = read_long(&fh,0);
	PR_DEBUG1("f_length = %08lX\n", f_length)

	f_check = read_long(&fh,(int)(f_length+sizeof(long)));
	PR_DEBUG1("f_check = %08lX\n", f_check)

	// compare f_length with f_check
	// if not valid reimageFS()
	// reimageFS() -> delete file, create file, copy ximage->xmem(ram)->FS
	// easiest way to do this is just to delete the file.
	// on the next boot, it gets re-created and copied to FS from xmem image.

	if (f_length != f_check)
	  {
	  fclose(&fh); // must close to delete.
     fs_sync();
	  fdelete(fhn);
	  PR_DEBUG("bad image! file deleted\n")
	  PR_DEBUG("reset (reboot) to reload initial image\n")
	  ErrorHandler(err_reload,-1);
	  }

	// now get ready to load forth from FS into ram, then boot forth from ram...

   ods=get_dataseg();

	PR_DEBUG1("DATASEG: %02x\n",ods)


   i=get_stackseg();

   PR_DEBUG1("STACKSEG: %02x\n",i)


   oss=get_segsize();

  	PR_DEBUG1("SEGSIZE: %02x\n",oss)
	PR_DEBUG1("physical address: %06lx\n",(long)forthbin)


	// baud rate is a parameter.
	// enables changing baud rate without recompiling forth.

   baud = BAUDRATE;
	// (divisor+1) for 19200 baud is stored in BIOS variable "freq_divider"
	divisor = (long)(freq_divider * 19200.0/(float)baud + 0.5) - 1L;
   // The counter divides by (n+1)

   br_div = (int)divisor;

	PR_DEBUG1("baud divisor: %d\n",br_div)
	PR_DEBUG1("BIOS freq_divider: %d\n",freq_divider)

	// change segsize...keeps
	// vars where they are & makes more ram available...
	// forth can now grow up from 0 to bottom of vars

	// put forth at next 4K boundary..._in_ the 48K of ram allocated earlier

	addr2=(((addr1 >> 12) + 1) * 0x1000)+FORTHORG;

	PR_DEBUG1("run code at pa: %06lx\n",(long)addr2)

	nds=(unsigned char)(addr2 >> 12);		// calculate new DATASEG

	// NOW LOAD addr1 with forth from FS

	read_file(&fh, addr1);

	fclose(&fh);		// CLOSE THE FILE!
	fs_sync();

	xmem2xmem(addr2,addr1+sizeof(long),(int)f_length);		// move forth from FS to ram
																			// apparently this moves high to low

	PR_DEBUG("starting forth...\n")
	PR_DEBUG("dummy line\n")
	PR_DEBUG("dummy line2\n")		// so we can see the 2nd previous line before d-c times out

	// structure old and new values for SEGSIZE & DATASEG
	// for in-line assy use...

	segsize_val=((oss & 0xF0) << 8)+ oss;		// (new * 0x100 + original)
	dataseg_val=(nds << 8)+ ods;



// boot forth image from ram...

#asm nodebug

   // stop ints & wdt before changing mem map!
	ld a,0x53
	ioi ld (WDTTR),a
	ld a,0x54
	ioi ld (WDTTR),a			; NO WATCHDOG!

	ioi ld a,(GCSR)
	and	0xFC					; no periodic ints!!!
	ioi	ld (GCSR),a


	ld	hl,(br_div)				; set up parameters

	ld bc,(segsize_val)
	ld de,(dataseg_val)

   // new MMU mapping
	ld A,b
	ioi	ld (SEGSIZE),A
	ld A,d
	ioi	ld (DATASEG),A

   // SP adjust to fully use 56K RAM

   ld (0xDFFC),SP	; stick SP in a safe place
   ld SP, 0xDFF0	; this is where the return address will go

   // START FORTH!

	call	FORTHORG				;with baudrate parameter in HL

   ld SP,(0xDFFC)	; restore original SP

   // restore MMU mapping
	ld A,e
	ioi	ld (DATASEG),A
	ld A,c
	ioi	ld (SEGSIZE),A

	ld (addr),hl				; get return value

#endasm



// forth exit returns to here so we can use the bios
// to flash a new image...
// or to restore the original image...

// return code:
//				0		no change to image (bye)
//				-1		re-install original image (forget)
//				addr	flash the new image (flash)


   SerialOpen(baud);

   // return value is next address in forth image
   //   if flash or forget used to quit
   //   or ZERO if BYE used to quit

   if (addr != 0)
     {
	  do	// prompt & response
	    {
   	 PR_SERIAL("\r\nWrite to flash (Y/N)?");
       do
         {
         c=SerialGetc();
         } while (c == -1);
       SerialPutc(c); // echo
		 if ((c == 'Y') || (c == 'N'))
		   {
		   ValidReply=TRUE;
		   }
		 else
		   {
		   ValidReply=FALSE;
		   PR_SERIAL("\r\nUPPERCASE ONLY\r\n");
		   }
       } while (!ValidReply);

     if ((c == 'Y') && (addr != -1))
       {
       PR_SERIAL("\r\nWriting Flash...\r\n");

 		 len16 = (addr-FORTHORG);	/* addr = base image + extension, the return value */

       // write new image to FS

		 addr2 = addr2 - sizeof(long);  // make room for length. ignore 4k page boundary.
       										  // can clobber whatever is before it under certain conditions.

       // put new length as long at beginning of buffer

		 root2xmem(addr2, &len16, sizeof(long));

       // new length also goes after image.

		 root2xmem((addr2+len16+sizeof(long)), &len16, sizeof(long));

       // delete file

       fs_sync();
		 rc = fdelete(fhn);
		 if (rc)
		   {
		   PR_SERIAL1("problem deleting old image (errno = %d)\r\n",errno);
		   }

		 fs_sync();

       // create file

       fs_set_lx(LX_2_USE, LX_2_USE);	// recommended before create...

		 rc = fcreate(&fh,fhn);
	    if (rc)
	      {
	      PR_SERIAL1("failed to create file! errno = %d\r\n",errno);
	      ErrorHandler(err_fcreatex,errno);
	      }

       // open for write

		 rc = fopen_wr(&fh,fhn);	/* now that it is created, open it for WRITE... */
	    if (rc)
	      {
	      PR_SERIAL("problem opening file!\r\n");
	      ErrorHandler(err_fopenX,rc);
	      }

       // write image

		 PR_SERIAL1("new image size = %04X\r\n", len16);

       // need a whole block for any number of bytes past BSIZE
       // 1 block min (len16 + 2*sizeof(long)):
       // 0008..1fff blk 1
       // 2000..3fff blk 2
       // 4000..5fff blk 3
       // 6000..7fff blk 4

		 nblks = ((len16+(2*sizeof(long)))/BSIZE)+1;
       PR_SERIAL1("# blocks = %d \r\n", nblks);

		 for(i=0;i<nblks;i++)
		   {
         ec = xmem2root(buffer,
         					(addr2 + (unsigned long)i*(unsigned long)BSIZE),
                        (unsigned int)BSIZE);
		   if (ec != 0) {
           PR_SERIAL1("Flash write failed! %d\r\n", ec);
           ErrorHandler(err_xmem2rootB,ec);
         }
		   write_file(&fh,buffer,BSIZE);
		   PR_SERIAL1("wrote buffer %d \r\n", i);
		   }

       // close file

		 fclose(&fh);

		 PR_SERIAL("\r\nFlash write done!\r\n");

       } // if (c == 'Y') && (addr != -1)
	  if ((c == 'Y') && (addr == -1))
       {		// addr = -1 : "forget" -> re-image on next boot
       fs_sync();
       rc = fdelete(fhn);		// already closed, as required
       if (rc)
         {
         PR_SERIAL1("problem deleting old image (errno = %d)\r\n",errno);
         }
       fs_sync();
       } // if (c == 'Y') && (addr == -1)
   } // if (addr != 0)


	PR_SERIAL("\r\nDone...\r\n");

   SerialClose();

} // end main


